import datetime
import json
from typing import Any

import factory
from django.utils import timezone
from factory.fuzzy import FuzzyDate

from canvas_sdk.v1.data import Referral, ReferralReport, ReferralReview


class ReferralFactory(factory.django.DjangoModelFactory[Referral]):
    """Factory for creating a Referral."""

    class Meta:
        model = Referral

    patient = factory.SubFactory("canvas_sdk.test_utils.factories.PatientFactory")
    note = factory.SubFactory("canvas_sdk.test_utils.factories.NoteFactory")
    service_provider = None  # Optional field
    clinical_question = factory.Faker("text", max_nb_chars=50)
    priority = factory.Faker("random_element", elements=["routine", "urgent", "stat"])
    include_visit_note = False
    notes = factory.Faker("paragraph")
    date_referred = factory.LazyFunction(lambda: timezone.now())
    forwarded = False
    internal_comment = factory.Faker("paragraph")
    internal_task_comment = None  # Optional field
    ignored = False
    task_ids = factory.LazyFunction(lambda: json.dumps([]))

    @factory.post_generation
    def assessments(self, create: Any, extracted: Any, **kwargs: Any) -> None:
        """Handle many-to-many relationship for assessments."""
        if not create:
            return

        if extracted:
            for assessment in extracted:
                self.assessments.add(assessment)


class ReferralReviewFactory(factory.django.DjangoModelFactory[ReferralReview]):
    """Factory for creating a ReferralReview."""

    class Meta:
        model = ReferralReview

    internal_comment = factory.Faker("paragraph")
    message_to_patient = factory.Faker("sentence", nb_words=20)
    status = factory.Faker("random_element", elements=["pending", "reviewed", "completed"])
    patient = factory.SubFactory("canvas_sdk.test_utils.factories.PatientFactory")
    patient_communication_method = factory.Faker(
        "random_element", elements=["email", "phone", "portal", "mail"]
    )


class ReferralReportFactory(factory.django.DjangoModelFactory[ReferralReport]):
    """Factory for creating a ReferralReport."""

    class Meta:
        model = ReferralReport

    originator = factory.SubFactory("canvas_sdk.test_utils.factories.CanvasUserFactory")
    review_mode = factory.Faker("random_element", elements=["IN", "OT"])
    assigned_by = factory.SubFactory("canvas_sdk.test_utils.factories.CanvasUserFactory")
    junked = False
    requires_signature = factory.Faker("boolean")
    assigned_date = factory.LazyFunction(lambda: timezone.now())
    team_assigned_date = None
    team = None
    patient = factory.SubFactory("canvas_sdk.test_utils.factories.PatientFactory")
    referral = factory.SubFactory(ReferralFactory)
    specialty = factory.Faker("job")
    review = None
    original_date = FuzzyDate(
        start_date=datetime.date.today() - datetime.timedelta(days=365),
        end_date=datetime.date.today(),
    )
    comment = factory.Faker("paragraph")
    priority = False



from functools import cached_property
from typing import TYPE_CHECKING

from django.db import models

from canvas_sdk.v1.data.base import Model

if TYPE_CHECKING:
    from canvas_sdk.v1.data import Patient, Staff


class CanvasUser(Model):
    """A class representing a Canvas User."""

    class Meta:
        db_table = "canvas_sdk_data_api_auth_user_001"

    email = models.EmailField(db_column="email")
    phone_number = models.CharField(db_column="phone_number", max_length=255)
    last_invite_date_time = models.DateTimeField(null=True, blank=True)
    is_portal_registered = models.BooleanField(default=False)
    is_staff = models.BooleanField(default=False)

    @cached_property
    def person_subclass(self) -> "Staff | Patient":
        """
        Return either the related Staff or Patient object.
        """
        return self.staff if self.is_staff else self.patient


__exports__ = ("CanvasUser",)



from datetime import datetime
from enum import Enum, StrEnum
from typing import Any, Self, cast
from uuid import UUID

from pydantic import model_validator
from pydantic_core import InitErrorDetails

from canvas_sdk.effects.base import EffectType, _BaseEffect
from canvas_sdk.effects.metadata import BaseMetadata
from canvas_sdk.v1.data import Task


class TaskStatus(Enum):
    """TaskStatus."""

    COMPLETED = "COMPLETED"
    CLOSED = "CLOSED"
    OPEN = "OPEN"


class AddTask(_BaseEffect):
    """
    An Effect that will create a Task in Canvas.
    """

    class LinkableObjectType(StrEnum):
        """Types of objects that can be linked to a Task."""

        REFERRAL = "REFERRAL"
        IMAGING = "IMAGING"

    class Meta:
        effect_type = EffectType.CREATE_TASK
        apply_required_fields = ("title",)

    id: str | UUID | None = None
    assignee_id: str | None = None
    team_id: str | None = None
    patient_id: str | None = None
    title: str | None = None
    due: datetime | None = None
    status: TaskStatus = TaskStatus.OPEN
    labels: list[str] = []
    linked_object_id: str | UUID | None = None
    linked_object_type: LinkableObjectType | None = None
    author_id: str | UUID | None = None

    @model_validator(mode="after")
    def check_needed_together_fields(self) -> Self:
        """Check that linked_object_id and linked_object_type are set together."""
        if self.linked_object_id is not None and self.linked_object_type is None:
            raise ValueError(
                "'linked_object_id' must be set with 'linked_object_type' if it is provided"
            )
        if self.linked_object_id is None and self.linked_object_type is not None:
            raise ValueError(
                "'linked_object_type' must be set with 'linked_object_id' if it is provided"
            )

        return self

    @property
    def values(self) -> dict[str, Any]:
        """The values for Task addition."""
        return {
            "id": str(self.id) if self.id else None,
            "patient": {"id": self.patient_id},
            "due": self.due.isoformat() if self.due else None,
            "assignee": {"id": self.assignee_id},
            "team": {"id": self.team_id},
            "title": self.title,
            "status": self.status.value,
            "labels": self.labels,
            "author_id": str(self.author_id) if self.author_id else None,
            "linked_object": {
                "id": str(self.linked_object_id) if self.linked_object_id else None,
                "type": self.linked_object_type.value if self.linked_object_type else None,
            },
        }


class AddTaskComment(_BaseEffect):
    """
    An Effect that will create a Task Comment on a Task.
    """

    class Meta:
        effect_type = EffectType.CREATE_TASK_COMMENT
        apply_required_fields = (
            "body",
            "task_id",
        )

    body: str | None = None
    task_id: str | UUID | None = None
    author_id: str | UUID | None = None

    @property
    def values(self) -> dict[str, Any]:
        """The values for adding a task comment."""
        return {
            "task": {"id": str(self.task_id) if self.task_id else None},
            "body": self.body,
            "author_id": str(self.author_id) if self.author_id else None,
        }


class UpdateTask(_BaseEffect):
    """
    An Effect that will update a Task in Canvas.
    """

    class Meta:
        effect_type = EffectType.UPDATE_TASK
        apply_required_fields = ("id",)

    id: str | None = None
    assignee_id: str | None = None
    team_id: str | None = None
    patient_id: str | None = None
    title: str | None = None
    due: datetime | None = None
    status: TaskStatus = TaskStatus.OPEN
    labels: list[str] = []

    @property
    def values(self) -> dict[str, Any]:
        """The values for adding a task comment."""
        value_dict: dict[str, Any] = {}
        # Only add the fields that have been explicitly set on the model (exclude_unset=True).
        # Oherwise, the effect interpreter will set values to null based on their defaults.
        set_fields = self.model_dump(exclude_unset=True)
        for field, val in set_fields.items():
            if field.endswith("_id"):
                value_dict[field.split("_")[0]] = {"id": val}
            elif field == "due" and val is not None:
                value_dict[field] = cast(datetime, val).isoformat()
            elif field == "status":
                value_dict[field] = cast(TaskStatus, val).value
            else:
                value_dict[field] = getattr(self, field)
        return value_dict


class TaskMetadata(BaseMetadata):
    """Task Metadata."""

    class Meta:
        effect_type = "TASK_METADATA"

    task_id: str

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        """Get the error details for the effect.
        If task_id is not found, return an error detail.
        """
        errors = super()._get_error_details(method)

        if not Task.objects.filter(id=self.task_id).exists():
            errors.append(
                self._create_error_detail(
                    "task_id",
                    f"Task with id: {self.task_id} does not exist.",
                    self.task_id,
                )
            )

        return errors


__exports__ = (
    "AddTask",
    "AddTaskComment",
    "TaskStatus",
    "TaskMetadata",
    "UpdateTask",
)



from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand


class StopMedicationCommand(_BaseCommand):
    """A class for managing a StopMedication command within a specific note."""

    class Meta:
        key = "stopMedication"

    # how do we make sure this is a valid medication_id for the patient?
    medication_id: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "medication"}
    )
    rationale: str | None = None


__exports__ = ("StopMedicationCommand",)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from canvas_sdk.caching.client import get_cache as get_cache_client
from canvas_sdk.utils.plugins import plugin_context
from settings import CANVAS_SDK_CACHE_TIMEOUT_SECONDS

if TYPE_CHECKING:
    from canvas_sdk.caching.base import Cache


@plugin_context
def get_cache(**kwargs: Any) -> Cache:
    """Get the cache client for plugins."""
    return get_cache_client(
        driver="plugins",
        prefix=kwargs["plugin_name"],
        max_timeout_seconds=CANVAS_SDK_CACHE_TIMEOUT_SECONDS,
    )


__exports__ = ("get_cache",)



from abc import ABC
from functools import cached_property
from typing import ClassVar

import sentry_sdk

from canvas_sdk.effects import Effect
from canvas_sdk.effects.simple_api import AcceptConnection, DenyConnection
from canvas_sdk.events import Event, EventType
from canvas_sdk.handlers.base import BaseHandler
from logger import log

from .exceptions import AuthenticationError, InvalidCredentialsError
from .tools import CaseInsensitiveMultiDict, separate_headers


class WebSocket:
    """WebSocket class for incoming requests to the WebSocketAPI."""

    def __init__(self, event: Event) -> None:
        self.channel = event.context["channel_name"]
        self.headers = CaseInsensitiveMultiDict(separate_headers(event.context["headers"]))
        self.api_key = self.headers.get("authorization")
        self.logged_in_user = (
            {
                "id": self.headers.get("canvas-logged-in-user-id", ""),
                "type": self.headers.get("canvas-logged-in-user-type", ""),
            }
            if "canvas-logged-in-user-id" in self.headers
            else None
        )


class WebSocketAPI(BaseHandler, ABC):
    """Abstract base class for WebSocket APIs."""

    RESPONDS_TO: ClassVar[list[str]] = [
        EventType.Name(EventType.SIMPLE_API_WEBSOCKET_AUTHENTICATE),
    ]

    @cached_property
    def websocket(self) -> WebSocket:
        """Return the WebSocket object for the event."""
        return WebSocket(self.event)

    def compute(self) -> list[Effect]:
        """Handle WebSocket authenticate event."""
        try:
            if self.event.type == EventType.SIMPLE_API_WEBSOCKET_AUTHENTICATE:
                return self._authenticate()
            else:
                raise AssertionError(f"Cannot handle event type {EventType.Name(self.event.type)}")
        except Exception as exception:
            log.exception(f"Error handling '{EventType.Name(self.event.type)}' event")

            sentry_sdk.capture_exception(exception)

            return [DenyConnection(message="Internal server error").apply()]

    def _authenticate(self) -> list[Effect]:
        """Authenticate the WebSocket request."""
        try:
            if self.authenticate():
                return [AcceptConnection().apply()]
            else:
                raise InvalidCredentialsError

        except AuthenticationError:
            return [DenyConnection(message="Unauthorized").apply()]

    def authenticate(self) -> bool:
        """Override to implement authentication."""
        return False


__exports__ = ("WebSocketAPI", "WebSocket")



import logging
import os
import sys
from pathlib import Path
from typing import Any, cast
from urllib import parse

from django.core.exceptions import ImproperlyConfigured
from dotenv import load_dotenv
from env_tools import env_to_bool

load_dotenv()

BASE_DIR = Path(__file__).resolve().parent.resolve()

FOURTEEN_DAYS = 60 * 60 * 24 * 14

ENV = os.getenv("ENV", "development")
IS_PRODUCTION = ENV == "production"
IS_PRODUCTION_CUSTOMER = env_to_bool("IS_PRODUCTION_CUSTOMER", IS_PRODUCTION)
IS_TESTING = env_to_bool("IS_TESTING", "pytest" in sys.argv[0] or sys.argv[0] == "-c")
IS_SCRIPT = env_to_bool("IS_SCRIPT", "plugin_runner.py" not in sys.argv[0])
PLUGIN_POOL_DEBUG = env_to_bool("PLUGIN_POOL_DEBUG")
CUSTOMER_IDENTIFIER = os.getenv("CUSTOMER_IDENTIFIER", "local")
HOSTNAME = os.getenv("HOSTNAME", "")
APP_NAME = os.getenv("APP_NAME")

if PLUGIN_POOL_DEBUG:
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("[%(process)d:%(thread)d] %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)

    pool_logger = logging.getLogger("psycopg.pool")
    pool_logger.setLevel(logging.DEBUG)
    pool_logger.addHandler(handler)

INTEGRATION_TEST_URL = os.getenv("INTEGRATION_TEST_URL")
INTEGRATION_TEST_CLIENT_ID = os.getenv("INTEGRATION_TEST_CLIENT_ID")
INTEGRATION_TEST_CLIENT_SECRET = os.getenv("INTEGRATION_TEST_CLIENT_SECRET")

GRAPHQL_ENDPOINT = os.getenv("GRAPHQL_ENDPOINT", "http://localhost:8000/plugins-graphql")
REDIS_ENDPOINT = os.getenv("REDIS_ENDPOINT", f"redis://{APP_NAME}-redis:6379")
CANVAS_SDK_PLUGINS_CACHE_LOCATION = os.getenv(
    "CANVAS_SDK_PLUGINS_CACHE_LOCATION", "plugin_io_plugins_cache"
)
CANVAS_SDK_CACHE_TIMEOUT_SECONDS = int(os.getenv("CANVAS_SDK_CACHE_TIMEOUT", FOURTEEN_DAYS))


METRICS_ENABLED = env_to_bool("PLUGINS_METRICS_ENABLED", not IS_SCRIPT)

INSTALLED_APPS = [
    "canvas_sdk.v1",
]

SECRET_KEY = os.getenv(
    "SECRET_KEY",
    "This value is the key to securing signed data – it is vital you keep this secure, or attackers could use it to generate their own signed values.",
)

# Use BigAutoField for Default Primary Key field type
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

CANVAS_SDK_DB_NAME = os.getenv("CANVAS_SDK_DB_NAME", "home-app")
CANVAS_SDK_DB_USERNAME = os.getenv("CANVAS_SDK_DB_USERNAME", "app")
CANVAS_SDK_DB_PASSWORD = os.getenv("CANVAS_SDK_DB_PASSWORD", "app")
CANVAS_SDK_DB_HOST = os.getenv("CANVAS_SDK_DB_HOST", "home-app-db")
CANVAS_SDK_DB_PORT = os.getenv("CANVAS_SDK_DB_PORT", "5432")
CANVAS_SDK_DB_URL = os.getenv("DATABASE_URL")

if IS_TESTING:
    CANVAS_SDK_DB_BACKEND = "sqlite3"
elif CANVAS_SDK_DB_URL:
    CANVAS_SDK_DB_BACKEND = "postgres"
else:
    # Default to sqlite3 for local development if no DATABASE_URL is set
    CANVAS_SDK_DB_BACKEND = "sqlite3" if IS_SCRIPT else "postgres"

PLUGIN_RUNNER_MAX_WORKERS = int(os.getenv("PLUGIN_RUNNER_MAX_WORKERS", 5))
CONN_HEALTH_CHECKS_ENABLED = env_to_bool("CONN_HEALTH_CHECKS_ENABLED", True)

# By default, allow a pool size that gives each worker 2 active connections
# and allow overriding via environment variable if necessary
PLUGIN_RUNNER_DATABASE_POOL_MAX = int(
    os.getenv(
        "PLUGIN_RUNNER_DATABASE_POOL_MAX",
        PLUGIN_RUNNER_MAX_WORKERS * 2,
    )
)

if CANVAS_SDK_DB_BACKEND == "postgres":
    db_config: dict[str, Any] = {
        "ENGINE": "django.db.backends.postgresql",
        "CONN_HEALTH_CHECKS": CONN_HEALTH_CHECKS_ENABLED,
        "OPTIONS": {
            "pool": {
                "min_size": 2,
                "max_size": PLUGIN_RUNNER_DATABASE_POOL_MAX,
            }
        },
    }

    if CANVAS_SDK_DB_URL:
        parsed_url = parse.urlparse(CANVAS_SDK_DB_URL)
        db_config.update(
            {
                "NAME": parsed_url.path[1:],
                "USER": os.getenv("CANVAS_SDK_DATABASE_ROLE"),
                "PASSWORD": os.getenv("CANVAS_SDK_DATABASE_ROLE_PASSWORD"),
                "HOST": parsed_url.hostname,
                "PORT": parsed_url.port,
            }
        )
    else:
        db_config.update(
            {
                "NAME": CANVAS_SDK_DB_NAME,
                "USER": CANVAS_SDK_DB_USERNAME,
                "PASSWORD": CANVAS_SDK_DB_PASSWORD,
                "HOST": CANVAS_SDK_DB_HOST,
                "PORT": CANVAS_SDK_DB_PORT,
            }
        )

    DATABASES = {"default": db_config}

elif CANVAS_SDK_DB_BACKEND == "sqlite3":
    SQLITE_DB_PATH = BASE_DIR / "canvas_db.sqlite3"
    SQLITE_WRITE_MODE_DATABASE = {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": str(SQLITE_DB_PATH),
    }

    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": f"file:{SQLITE_DB_PATH}?mode=ro",
            "OPTIONS": {"uri": True},
        }
    }


else:
    raise ImproperlyConfigured(
        "Unsupported database backend specified in 'CANVAS_SDK_DB_BACKEND' setting. "
        "Supported values are 'postgres' and 'sqlite3'."
    )

AWS_ACCESS_KEY_ID = os.getenv("AWS_ACCESS_KEY_ID", "")
AWS_SECRET_ACCESS_KEY = os.getenv("AWS_SECRET_ACCESS_KEY", "")
AWS_REGION = os.getenv("AWS_REGION", "us-west-2")
MEDIA_S3_BUCKET_NAME = os.getenv("MEDIA_S3_BUCKET_NAME", "canvas-client-media")

# ONTOLOGIES_SIGNING_KEY = os.getenv("ONTOLOGIES_SIGNING_KEY", "")
PLUGIN_RUNNER_SIGNING_KEY = os.getenv("PLUGIN_RUNNER_SIGNING_KEY", "")
# SCIENCE_SIGNING_KEY = os.getenv("SCIENCE_SIGNING_KEY", "")

PLUGIN_DIRECTORY = os.getenv(
    "PLUGIN_DIRECTORY",
    (
        "/plugin-runner/custom-plugins"
        if IS_PRODUCTION
        else (BASE_DIR / "plugin_runner/tests/data/plugins").as_posix()
        if IS_TESTING
        else (BASE_DIR / "custom-plugins").as_posix()
    ),
)
PLUGINS_PUBSUB_CHANNEL = os.getenv("PLUGINS_PUBSUB_CHANNEL", default="plugins")
CHANNEL_NAME = f"{CUSTOMER_IDENTIFIER}:{PLUGINS_PUBSUB_CHANNEL}"
MANIFEST_FILE_NAME = "CANVAS_MANIFEST.json"

SECRETS_FILE_NAME = "SECRETS.json"

SENTRY_DSN = os.getenv("SENTRY_DSN")


if IS_SCRIPT:
    CACHES = {
        "default": {
            "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        },
        "plugins": {
            "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
            "KEY_PREFIX": CUSTOMER_IDENTIFIER,
            "LOCATION": "plugins_cache",
            "TIMEOUT": CANVAS_SDK_CACHE_TIMEOUT_SECONDS,
        },
    }
else:
    CACHES = {
        "plugins": {
            "BACKEND": "django.core.cache.backends.db.DatabaseCache",
            "KEY_PREFIX": CUSTOMER_IDENTIFIER,
            "LOCATION": CANVAS_SDK_PLUGINS_CACHE_LOCATION,
            "TIMEOUT": CANVAS_SDK_CACHE_TIMEOUT_SECONDS,
        }
    }

LOGSTASH_HOST = os.getenv("PLUGINS_LOGSTASH_URL")
LOGSTASH_PORT = (
    int(cast(str, os.getenv("PLUGINS_LOGSTASH_PORT")))
    if os.getenv("PLUGINS_LOGSTASH_PORT")
    else None
)
LOGSTASH_PROTOCOL = os.getenv("PLUGINS_LOGSTASH_PROTOCOL", "logger.logstash.HttpTransport")



from canvas_sdk.commands.commands.review.base import _BaseReview
from canvas_sdk.v1.data import LabReport


class LabReviewCommand(_BaseReview):
    """A class for managing a Lab Review command within a specific note."""

    class Meta:
        key = "labReview"
        model = LabReport


__exports__ = ("LabReviewCommand",)



from pydantic import Field
from pydantic_core import InitErrorDetails

from canvas_sdk.commands.base import _BaseCommand as BaseCommand
from canvas_sdk.commands.constants import CodeSystems, Coding


class InstructCommand(BaseCommand):
    """A class for managing an Instruct command within a specific note."""

    class Meta:
        key = "instruct"

    coding: Coding | None = Field(default=None, json_schema_extra={"commands_api_name": "instruct"})
    comment: str | None = None

    def _get_error_details(self, method: str) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        if (
            self.coding
            and self.coding["system"] != CodeSystems.SNOMED
            and self.coding["system"] != CodeSystems.UNSTRUCTURED
        ):
            message = f"The 'coding.system' field must be '{CodeSystems.SNOMED}' or '{CodeSystems.UNSTRUCTURED}'."
            errors.append(self._create_error_detail("value", message, self.coding))

        return errors


__exports__ = ("InstructCommand",)



import datetime

import factory
from factory.fuzzy import FuzzyDate

from canvas_sdk.test_utils.factories.facility import FacilityFactory
from canvas_sdk.test_utils.factories.user import CanvasUserFactory
from canvas_sdk.v1.data import Patient, PatientAddress, PatientFacilityAddress


class PatientAddressFactory(factory.django.DjangoModelFactory[PatientAddress]):
    """Factory for creating a PatientAddress."""

    class Meta:
        model = PatientAddress

    line1 = "1234 Main Street"
    line2 = "Apt 3"
    city = "San Francisco"
    district = "Sunset"
    state_code = "CA"
    postal_code = "94112"
    country = "USA"


class PatientFactory(factory.django.DjangoModelFactory[Patient]):
    """Factory for creating a Patient."""

    class Meta:
        model = Patient

    birth_date = FuzzyDate(
        start_date=datetime.date.today() - datetime.timedelta(days=100 * 365),
        end_date=datetime.date.today() - datetime.timedelta(days=10 * 365),
    )
    first_name = factory.Faker("first_name")
    middle_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    addresses = factory.RelatedFactory(PatientAddressFactory, "patient")
    user = factory.SubFactory(CanvasUserFactory)


class PatientFacilityAddressFactory(factory.django.DjangoModelFactory[PatientFacilityAddress]):
    """Factory for creating a PatientFacilityAddress."""

    class Meta:
        model = PatientFacilityAddress

    facility = factory.SubFactory(FacilityFactory)
    patient = factory.SubFactory(PatientFactory)
    room_number = factory.Faker("building_number")



from functools import cached_property

from django.db import models

from canvas_sdk.v1.data.base import IdentifiableModel


class ServiceProvider(IdentifiableModel):
    """ServiceProvider."""

    class Meta:
        db_table = "canvas_sdk_data_data_integration_serviceprovider_001"

    first_name = models.CharField(max_length=512)
    # organizations won't have a last name
    last_name = models.CharField(max_length=512, default="", blank=True)
    business_fax = models.CharField(max_length=512, null=True, blank=True)
    business_phone = models.CharField(max_length=512, null=True, blank=True)
    business_address = models.CharField(max_length=512, null=True, blank=True)
    specialty = models.CharField(max_length=512)
    practice_name = models.CharField(max_length=512, null=True, blank=True)
    notes = models.TextField(default="", null=True, blank=True)

    @property
    def full_name(self) -> str:
        """Service provider full name."""
        return f"{self.first_name} {self.last_name}"

    @cached_property
    def full_name_and_specialty(self) -> str:
        """Service provider full name and specialty."""
        name_components: list[str] = []

        # Note 1: if firstName is (TBD) then insert at the end instead of the beginning
        if self.first_name != "(TBD)":
            name_components.append(self.first_name)

        if self.first_name != self.last_name:
            name_components.append(self.last_name)

        if self.practice_name and self.practice_name != "(TBD)":
            name_components.append(f"({self.practice_name}),")

        if self.specialty not in [self.first_name, self.last_name, self.practice_name]:
            name_components.append(self.specialty)

        # see Note 1
        if self.first_name == "(TBD)":
            name_components.append(self.first_name)

        return " ".join(name_components)


__exports__ = ("ServiceProvider",)



from typing import Literal
from uuid import UUID

from django.db.models.query_utils import Q
from pydantic import Field
from pydantic_core import InitErrorDetails

from canvas_sdk.commands.base import _BaseCommand as BaseCommand
from canvas_sdk.commands.base import _SendableCommandMixin
from canvas_sdk.v1.data.lab import LabPartner, LabPartnerTest


class LabOrderCommand(_SendableCommandMixin, BaseCommand):
    """A class for managing a Lab Order command within a specific note."""

    class Meta:
        key = "labOrder"

    lab_partner: UUID | str | None = None
    tests_order_codes: list[str] = Field(
        default=[], json_schema_extra={"commands_api_name": "tests"}
    )
    ordering_provider_key: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "ordering_provider"}
    )
    diagnosis_codes: list[str] = Field(
        default=[], json_schema_extra={"commands_api_name": "diagnosis"}
    )
    fasting_required: bool = False
    comment: str | None = None

    def _get_error_details(
        self, method: Literal["originate", "edit", "delete", "commit", "enter_in_error", "send"]
    ) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        lab_partner_obj = None
        if self.lab_partner:
            query = {}
            try:
                UUID(str(self.lab_partner))
            except ValueError:
                query["name"] = self.lab_partner
            else:
                query["id"] = self.lab_partner

            lab_partner_obj = (
                LabPartner.objects.filter(**query)
                .values("id", "dbid", "electronic_ordering_enabled")
                .first()
            )

            if not lab_partner_obj:
                errors.append(
                    self._create_error_detail(
                        "value",
                        f"lab partner with Id or Name {self.lab_partner} not found",
                        self.lab_partner,
                    )
                )

        if self.tests_order_codes:
            if not lab_partner_obj:
                errors.append(
                    self._create_error_detail(
                        "value", "lab partner is required to find tests", self.tests_order_codes
                    )
                )
            else:
                uuid_tests = []
                order_code_tests = []

                for code in self.tests_order_codes:
                    try:
                        uuid_tests.append(UUID(str(code)))
                    except ValueError:
                        order_code_tests.append(code)

                tests = LabPartnerTest.objects.filter(
                    Q(order_code__in=order_code_tests) | Q(id__in=uuid_tests),
                    lab_partner_id=lab_partner_obj["dbid"],
                ).values_list("id", "order_code")

                if tests.count() != len(self.tests_order_codes):
                    missing_tests = [
                        test
                        for test in self.tests_order_codes
                        if not any(
                            test == str(test_id) or test == order_code
                            for test_id, order_code in tests
                        )
                    ]

                    errors.append(
                        self._create_error_detail(
                            "value",
                            f"tests with order codes {missing_tests} not found",
                            self.tests_order_codes,
                        )
                    )

        return errors


__exports__ = ("LabOrderCommand",)



from typing import TYPE_CHECKING, Self

from django.db import models

from canvas_sdk.v1.data.base import (
    IdentifiableModel,
    TimestampedModel,
    ValueSetTimeframeLookupQuerySet,
)
from canvas_sdk.v1.data.coding import Coding
from canvas_sdk.value_set.value_set import CodeConstants

if TYPE_CHECKING:
    from canvas_sdk.value_set.value_set import ValueSet


class BillingLineItemQuerySet(ValueSetTimeframeLookupQuerySet):
    """A class that adds functionality to filter BillingLineItem objects."""

    def find(self, value_set: type["ValueSet"]) -> Self:
        """
        This method is overridden to use for BillingLineItem CPT codes.
        The codes are saved as string values in the BillingLineItem.cpt field,
        which differs from other coding models.
        """
        values_dict = value_set.values
        return self.filter(cpt__in=values_dict.get(CodeConstants.HCPCS, []))


class BillingLineItemStatus(models.TextChoices):
    """Billing line item status."""

    ACTIVE = "active", "Active"
    REMOVED = "removed", "Removed"


class BillingLineItem(TimestampedModel, IdentifiableModel):
    """BillingLineItem."""

    class Meta:
        db_table = "canvas_sdk_data_api_billinglineitem_001"

    objects = models.Manager().from_queryset(BillingLineItemQuerySet)()

    note = models.ForeignKey(
        "v1.Note",
        on_delete=models.DO_NOTHING,
        related_name="billing_line_items",
        null=True,
    )
    patient = models.ForeignKey(
        "v1.Patient",
        on_delete=models.DO_NOTHING,
        related_name="billing_line_items",
        null=True,
    )
    cpt = models.CharField(max_length=10)
    charge = models.DecimalField(decimal_places=2, max_digits=8)
    description = models.CharField(max_length=255)
    units = models.IntegerField()
    command_type = models.CharField(max_length=50)
    command_id = models.IntegerField()
    status = models.CharField(choices=BillingLineItemStatus.choices, max_length=20)


class BillingLineItemModifier(Coding):
    """BillingLineItemModifier."""

    class Meta:
        db_table = "canvas_sdk_data_api_billinglineitemmodifier_001"

    line_item = models.ForeignKey(
        "v1.BillingLineItem",
        on_delete=models.DO_NOTHING,
        related_name="modifiers",
        null=True,
    )


__exports__ = (
    "BillingLineItemStatus",
    "BillingLineItem",
    "BillingLineItemModifier",
)



from enum import StrEnum
from typing import Any
from uuid import UUID

from canvas_sdk.effects import EffectType, _BaseEffect


class CalendarType(StrEnum):
    """Calendar type."""

    Clinic = "Clinic"
    Administrative = "Admin"


class CreateCalendar(_BaseEffect):
    """Effect to create a Calendar."""

    class Meta:
        effect_type = EffectType.CALENDAR__CREATE

    id: str | UUID | None = None
    provider: str | UUID
    type: CalendarType
    description: str | None = None

    @property
    def values(self) -> dict[str, Any]:
        """The Calendar's values."""
        return {
            "id": self.id,
            "provider": self.provider,
            "type": self.type,
            "description": self.description,
        }


__exports__ = (
    "CreateCalendar",
    "CalendarType",
)



from canvas_sdk.effects.appointments_metadata.appointments_metadata_create_form import (
    AppointmentsMetadataCreateFormEffect,
    FormField,
    InputType,
)
from canvas_sdk.effects.appointments_metadata.base import AppointmentsMetadata

__all__ = __exports__ = (
    "FormField",
    "InputType",
    "AppointmentsMetadata",
    "AppointmentsMetadataCreateFormEffect",
)



from typing import cast

from django.db.models.expressions import Subquery
from pydantic_core import InitErrorDetails

from canvas_sdk.commands.commands.prescribe import PrescribeCommand
from canvas_sdk.v1.data import Medication, Note


class RefillCommand(PrescribeCommand):
    """A class for managing a Refill command within a specific note."""

    class Meta:
        key = "refill"

    def _get_error_details(self, method: str) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        if self.fdb_code:
            subquery = Subquery(
                Note.objects.filter(id=cast(str, self.note_uuid)).values("patient_id")[:1]
            )
            if (
                not Medication.objects.active()
                .filter(codings__code=self.fdb_code, patient=subquery)
                .exists()
            ):
                errors.append(
                    self._create_error_detail(
                        "value",
                        f"Medication with fdb_code {self.fdb_code} does not exist.",
                        self.fdb_code,
                    )
                )

        return errors


__exports__ = ("RefillCommand",)



import json
from datetime import datetime
from pathlib import Path
from typing import Any

TOKENS_PATH = Path.home() / ".canvas" / "tokens.json"
TOKENS_PATH.parent.mkdir(parents=True, exist_ok=True)


def _load_tokens() -> dict:
    if not TOKENS_PATH.exists():
        return {}
    with TOKENS_PATH.open("r") as f:
        return json.load(f)


def _save_tokens(tokens: dict) -> None:
    with TOKENS_PATH.open("w") as f:
        json.dump(tokens, f, indent=2)


def _set_value(key: str, value: Any) -> None:
    """Set a value in the token storage."""
    tokens = _load_tokens()
    tokens[key] = value
    _save_tokens(tokens)


def _get_value(key: str) -> Any | None:
    """Get value from the token storage."""
    tokens = _load_tokens()
    return tokens.get(key)


def _delete_value(key: str) -> None:
    """Delete a value from the token storage."""
    tokens = _load_tokens()
    tokens.pop(key, None)
    _save_tokens(tokens)


def set_token(host_token_key: str, token: str, exp: datetime) -> None:
    """Set a token with an expiry date in the token storage."""
    _set_value(host_token_key, {"token": token, "exp_date": exp.isoformat()})


def get_token(host_token_key: str) -> str | None:
    """Get a token from the token storage."""
    value = _get_value(host_token_key)

    if not value:
        return None
    try:
        exp = value["exp_date"]
        if datetime.fromisoformat(exp) <= datetime.now():
            return None
    except ValueError:
        return None

    return value["token"]


def delete_token(host_token_key: str) -> None:
    """Delete a token from the token storage."""
    _delete_value(host_token_key)



from django.db import models

from canvas_sdk.v1.data.base import Model


class CDMCodeSystem(models.TextChoices):
    """CDMCodeSystem."""

    INTERNAL = "INTERNAL", "Internal"
    CPT = "CPT", "CPT"


class ChargeDescriptionMaster(Model):
    """Charge Description Master."""

    class Meta:
        db_table = "canvas_sdk_data_quality_and_revenue_chargedescriptionmaster_001"

    cpt_code = models.CharField(max_length=10)
    name = models.CharField(max_length=17000)
    short_name = models.CharField(max_length=17000)
    charge_amount = models.DecimalField(max_digits=8, decimal_places=2)
    effective_date = models.DateField()
    end_date = models.DateField(null=True)
    code_system = models.CharField(choices=CDMCodeSystem.choices, max_length=10)
    ndc_code = models.CharField(max_length=255)


__exports__ = ("ChargeDescriptionMaster",)



from django.db import models

from canvas_sdk.v1.data.base import IdentifiableModel, TimestampedModel


class ClaimDiagnosisCode(TimestampedModel, IdentifiableModel):
    """ClaimDiagnosisCode model for claim diagnosis codes."""

    class Meta:
        db_table = "canvas_sdk_data_quality_and_revenue_claimdiagnosiscode_001"
        ordering = ["rank"]

    claim = models.ForeignKey("v1.Claim", on_delete=models.CASCADE, related_name="diagnosis_codes")
    rank = models.IntegerField()
    code = models.CharField(max_length=20, blank=True)
    display = models.CharField(max_length=1000, blank=True)


__exports__ = ("ClaimDiagnosisCode",)



from canvas_sdk.effects import Effect


def normalize_effects(effects: Effect | list[Effect] | None) -> list[Effect]:
    """Normalize effects to a list of Effect instances."""
    if effects is None:
        return []
    if isinstance(effects, Effect):
        return [effects]
    if isinstance(effects, list):
        return [e for e in effects if isinstance(e, Effect)]


__exports__ = ()



from .task import AddTask, AddTaskComment, TaskMetadata, TaskStatus, UpdateTask

__all__ = __exports__ = (
    "AddTask",
    "AddTaskComment",
    "TaskStatus",
    "TaskMetadata",
    "UpdateTask",
)



from datetime import datetime
from enum import Enum

from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand


class UpdateGoalCommand(_BaseCommand):
    """A class for managing an UpdateGoal command within a specific note."""

    class Meta:
        key = "updateGoal"

    class AchievementStatus(Enum):
        IN_PROGRESS = "in-progress"
        IMPROVING = "improving"
        WORSENING = "worsening"
        NO_CHANGE = "no-change"
        ACHIEVED = "achieved"
        SUSTAINING = "sustaining"
        NOT_ACHIEVED = "not-achieved"
        NO_PROGRESS = "no-progress"
        NOT_ATTAINABLE = "not-attainable"

    class Priority(Enum):
        HIGH = "high-priority"
        MEDIUM = "medium-priority"
        LOW = "low-priority"

    goal_id: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "goal_statement"}
    )
    due_date: datetime | None = None
    achievement_status: AchievementStatus | None = None
    priority: Priority | None = None
    progress: str | None = None


__exports__ = ("UpdateGoalCommand",)



from canvas_sdk.effects.calendar.create_calendar import CalendarType, CreateCalendar
from canvas_sdk.effects.calendar.create_event import CreateEvent, EventRecurrence

__all__ = __exports__ = ("CreateCalendar", "CalendarType", "CreateEvent", "EventRecurrence")



from django.db import models

from canvas_sdk.v1.data.base import AuditedModel, IdentifiableModel


class MedicationStatement(AuditedModel, IdentifiableModel):
    """MedicationStatement."""

    class Meta:
        db_table = "canvas_sdk_data_api_medicationstatement_001"

    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="medication_statements"
    )
    note = models.ForeignKey(
        "v1.Note", on_delete=models.DO_NOTHING, related_name="medication_statements"
    )
    medication = models.ForeignKey(
        "v1.Medication",
        on_delete=models.DO_NOTHING,
        related_name="medication_statements",
        null=True,
    )
    indications = models.ManyToManyField(
        "v1.Assessment",
        related_name="treatments_stated",
        db_table="canvas_sdk_data_api_medicationstatement_indications_001",
    )
    start_date_original_input = models.CharField(max_length=255, default="")
    start_date = models.DateField(default=None, null=True)
    end_date_original_input = models.CharField(max_length=255, default="")
    end_date = models.DateField(default=None, null=True)
    dose_quantity = models.FloatField(null=True)
    dose_form = models.CharField(max_length=255, default="")
    dose_route = models.CharField(max_length=255, default="")
    dose_frequency = models.FloatField(null=True)
    dose_frequency_interval = models.CharField(max_length=255, default="")
    sig_original_input = models.CharField(max_length=1000, default="")


__exports__ = ("MedicationStatement",)



#!/usr/bin/env uv run

import json
import logging
import re
import sys
import traceback
from collections import defaultdict
from contextlib import redirect_stderr, redirect_stdout
from glob import glob
from io import StringIO
from pathlib import Path
from textwrap import indent

import httpretty

from canvas_generated.messages.events_pb2 import PLUGIN_CREATED
from canvas_generated.messages.events_pb2 import Event as EventRequest
from canvas_sdk.events.base import Event
from canvas_sdk.protocols.base import BaseProtocol
from plugin_runner.plugin_runner import LOADED_PLUGINS, load_or_reload_plugin

ORIGINAL_PATH = sys.path.copy()


def enable_httpretty() -> None:
    """
    Prevent any HTTP requests from being made.
    """
    httpretty.enable(verbose=True, allow_net_connect=False)

    for method in [
        httpretty.GET,
        httpretty.PUT,
        httpretty.POST,
        httpretty.DELETE,
        httpretty.HEAD,
        httpretty.PATCH,
        httpretty.OPTIONS,
        httpretty.CONNECT,
    ]:
        httpretty.register_uri(method, re.compile(r".*"), body="{}")


def disable_httpretty() -> None:
    """
    Allow normal HTTP activity.
    """
    httpretty.disable()
    httpretty.reset()


def main() -> None:
    """
    Iterate through all of our plugins to verify they work with this version of the SDK.
    """
    if len(sys.argv) < 2:
        print("You must specify a base path.")
        sys.exit(1)

    base_path = sys.argv[1]

    for folder in glob(f"{base_path}/**/*"):
        path = Path(folder)

        if "{{" in folder:
            print(f"Ignoring templated path: {folder}")
            continue

        if not path.is_dir():
            continue

        if not (path / "CANVAS_MANIFEST.json").exists():
            print(f"Missing CANVAS_MANIFEST.json in {folder}")
            continue

        sys.path = ORIGINAL_PATH.copy()

        # this makes plugin imports work and mirrors what the plugin_runner does
        sys.path.append((Path(".") / path.parent).as_posix())

        print(path)

        output = StringIO()

        plugin_runner_logger = logging.getLogger("plugin_runner_logger")
        old_handlers = plugin_runner_logger.handlers.copy()
        plugin_runner_logger.handlers = [logging.StreamHandler(output)]

        with redirect_stdout(output), redirect_stderr(output):
            success = load_or_reload_plugin(path)

        output_string = output.getvalue()

        if not success:
            print()
            print(indent(output_string, prefix="    ").rstrip())
            print()

        plugin_runner_logger.handlers = old_handlers

    enable_httpretty()

    for key, plugin in LOADED_PLUGINS.items():
        Klass: type[BaseProtocol] = plugin["class"]

        if not hasattr(Klass, "compute") or not callable(Klass.compute):
            print(f"Skipping {key} due to no compute method")

        event_request = EventRequest(
            type=PLUGIN_CREATED,
            context=json.dumps(
                {
                    "key": "test",
                    "fields": {
                        "coding": {"value": "test"},
                        "goal_statement": {"value": "test"},
                        "questionnaire": {"text": "test"},
                        "medication": {"extra": {"coding": "test"}},
                        "ordering_provider": {"value": "test"},
                        "text": "test",
                    },
                    "method": "GET",
                    "note_id": 5,
                    "note": {
                        "uuid": "b2bd0672-ab36-4eca-b105-884f976ae810",
                        "id": "b2bd0672-ab36-4eca-b105-884f976ae810",
                    },
                    "path": "https://test",
                    "patient_id": 5,
                    "patient": {"id": 5},
                    "results": [],
                    "search_term": "test",
                    "state": "LKD",
                }
            ),
            target="b2bd0672-ab36-4eca-b105-884f976ae810",
            target_type="Patient",
        )
        event = Event(event_request=event_request)

        try:
            klass = Klass(
                event=event,
                secrets=defaultdict(lambda: "test", {"sentinel": True}),
                environment=defaultdict(lambda: "test", {"sentinel": True}),
            )
        except TypeError as e:
            print(f"Failed to instantiate {key}:", e)
            continue

        plugin_runner_logger = logging.getLogger("plugin_runner_logger")
        old_handlers = plugin_runner_logger.handlers.copy()
        plugin_runner_logger.handlers = [logging.StreamHandler(output)]

        output = StringIO()

        error: Exception | None = None

        try:
            with redirect_stdout(output), redirect_stderr(output):
                result = klass.compute()
        except TypeError as e:
            error = e
        except KeyError as e:
            error = e
        except Exception as e:
            error = e

        plugin_runner_logger.handlers = old_handlers

        result_count = len(result)
        request_count = len(httpretty.latest_requests())

        print(f"{key}: {result_count}/{request_count}")

        if error:
            print()

            # abbreviate these, they come from not being able to access the database
            if "fake_getaddrinfo" in str(error):
                print(indent(str(error), prefix="    ").rstrip())
            else:
                print(indent("".join(traceback.format_exception(error)), prefix="    ").rstrip())

            print()

        output_string = output.getvalue()

        if output_string:
            print()
            print(indent(output_string, prefix="    ").rstrip())
            print()

    disable_httpretty()


if __name__ == "__main__":
    main()



import factory
from factory.fuzzy import FuzzyInteger

from canvas_sdk.v1.data import ClaimDiagnosisCode


class ClaimDiagnosisCodeFactory(factory.django.DjangoModelFactory[ClaimDiagnosisCode]):
    """Factory for creating a ClaimDiagnosisCode."""

    class Meta:
        model = ClaimDiagnosisCode

    rank = FuzzyInteger(1, 12)  # ICD-10 allows up to 12 diagnosis codes on a claim
    code = factory.Faker("bothify", text="?##.##")  # ICD-10 format like A01.23
    display = factory.Faker("sentence", nb_words=4)  # Human readable diagnosis description
    claim = factory.SubFactory("canvas_sdk.test_utils.factories.ClaimFactory")



from django.db import models
from django.db.models import QuerySet

from canvas_sdk.v1.data.base import TimestampedModel


class AbstractLineItemQuerySet(models.QuerySet):
    """AbstractLineItemQuerySet."""

    def active(self) -> QuerySet:
        """Filter out line items that have been entered in error."""
        return self.filter(entered_in_error__isnull=True)


class AbstractLineItemTransaction(TimestampedModel):
    """Abstract class with common properties for both payments and adjustments."""

    class Meta:
        abstract = True

    objects = AbstractLineItemQuerySet.as_manager()

    posting = models.ForeignKey(
        "v1.BasePosting", related_name="%(class)ss", on_delete=models.PROTECT
    )
    billing_line_item = models.ForeignKey(
        "v1.BillingLineItem", related_name="%(class)ss", on_delete=models.PROTECT
    )
    amount = models.DecimalField(max_digits=8, decimal_places=2)


class NewLineItemPayment(AbstractLineItemTransaction):
    """Subclass that represents a payment on a billing line item."""

    class Meta:
        db_table = "canvas_sdk_data_quality_and_revenue_newlineitempayment_001"

    charged = models.DecimalField(max_digits=8, decimal_places=2)


class AbstractLineItemAdjustment(AbstractLineItemTransaction):
    """Abstract subclass with common fields for both adjustment and transfer."""

    class Meta:
        abstract = True

    code = models.CharField(max_length=8)
    group = models.CharField(max_length=8)

    deviated_from_posting_ruleset = models.BooleanField()


class NewLineItemAdjustment(AbstractLineItemAdjustment):
    """Subclass for billing line item adjustments."""

    class Meta:
        db_table = "canvas_sdk_data_quality_and_revenue_newlineitemadjustment_001"

    write_off = models.BooleanField()


class LineItemTransfer(AbstractLineItemAdjustment):
    """Subclass for billing line item balance transfer to other coverages or patient."""

    class Meta:
        db_table = "canvas_sdk_data_quality_and_revenue_lineitemtransfer_001"

    transfer_to = models.ForeignKey(
        "v1.ClaimCoverage", related_name="transfers", on_delete=models.PROTECT, null=True
    )
    transfer_to_patient = models.BooleanField(default=False)


__exports__ = (
    "NewLineItemPayment",
    "NewLineItemAdjustment",
    "LineItemTransfer",
)



from typing import Any

from canvas_sdk.effects.base import EffectType, _BaseEffect
from canvas_sdk.effects.group import Group


class PatientChartGroup(_BaseEffect):
    """
    An Effect that groups chart items by name and priority.
    """

    class Meta:
        effect_type = EffectType.PATIENT_CHART__GROUP_ITEMS

    items: dict[str, Group]

    @property
    def values(self) -> dict[str, Any]:
        """The chart items."""
        return {
            "items": [item.to_dict() for item in self.items.values() if isinstance(item, Group)]
        }

    @property
    def effect_payload(self) -> dict[str, Any]:
        """The payload of the effect."""
        return {"data": self.values}


__exports__ = ("PatientChartGroup",)



from abc import ABC
from typing import Any, Literal
from uuid import UUID

from pydantic_core import InitErrorDetails

from canvas_sdk.effects import EffectType, _BaseEffect
from canvas_sdk.v1.data import PracticeLocation


class BaseFaxEffect(_BaseEffect, ABC):
    """Base class for fax effects."""

    class Meta:
        effect_type = EffectType.UNKNOWN_EFFECT

    recipient_name: str
    recipient_fax_number: str
    include_coversheet: bool = False
    subject: str | None = None
    comment: str | None = None
    location_id: str | UUID | None = None

    def _get_error_details(self, method: Literal["apply"]) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        if self.include_coversheet:
            if not self.subject:
                errors.append(
                    self._create_error_detail(
                        "value",
                        "subject is required when include_coversheet is True",
                        self.subject,
                    )
                )
            if not self.comment:
                errors.append(
                    self._create_error_detail(
                        "value",
                        "comment is required when include_coversheet is True",
                        self.comment,
                    )
                )
            if self.location_id:
                if not PracticeLocation.objects.filter(id=self.location_id).exists():
                    errors.append(
                        self._create_error_detail(
                            "value",
                            f"Practice Location {self.location_id} does not exist",
                            self.location_id,
                        )
                    )
            else:
                errors.append(
                    self._create_error_detail(
                        "value",
                        "location_id is required when include_coversheet is True",
                        self.location_id,
                    )
                )

        return errors

    @property
    def values(self) -> dict[str, Any]:
        """Return the values of the fax effect."""
        return {
            "recipient_name": self.recipient_name,
            "recipient_fax_number": self.recipient_fax_number,
            "include_coversheet": self.include_coversheet,
            "subject": self.subject,
            "comment": self.comment,
            "location_id": str(self.location_id) if self.location_id else None,
        }


__exports__ = ()



from canvas_sdk.effects.note.appointment import (
    AddAppointmentLabel,
    Appointment,
    RemoveAppointmentLabel,
    ScheduleEvent,
)
from canvas_sdk.effects.note.base import AppointmentIdentifier
from canvas_sdk.effects.note.note import Note

__all__ = __exports__ = (
    "AppointmentIdentifier",
    "Note",
    "Appointment",
    "ScheduleEvent",
    "AddAppointmentLabel",
    "RemoveAppointmentLabel",
)



import factory

from canvas_sdk.v1.data import Organization


class OrganizationFactory(factory.django.DjangoModelFactory[Organization]):
    """Factory for creating Organization."""

    class Meta:
        model = Organization

    short_name = "Canvas Medical"
    full_name = "Canvas Medical, Inc."
    logo_url = "https://s3-us-west-2.amazonaws.com/canvas-pcp-assets/canvas_logo_large.png"
    background_image_url = "https://s3-us-west-2.amazonaws.com/canvas-pcp-assets/santa-cruz-low.jpg"
    background_gradient = (
        "linear-gradient(to top, rgb(231, 201, 154) 0%, rgb(220, 210, 192) 29%, "
        "rgb(156, 175, 191) 90%, rgb(147, 166, 189) 100%)"
    )
    tax_id = "123456"



from .allergy_intolerance import AllergyIntolerance, AllergyIntoleranceCoding
from .appointment import (
    Appointment,
    AppointmentExternalIdentifier,
    AppointmentLabel,
    AppointmentMetadata,
)
from .assessment import Assessment
from .banner_alert import BannerAlert
from .billing import BillingLineItem, BillingLineItemModifier
from .business_line import BusinessLine
from .care_team import CareTeamMembership, CareTeamRole
from .charge_description_master import ChargeDescriptionMaster
from .claim import (
    Claim,
    ClaimComment,
    ClaimCoverage,
    ClaimLabel,
    ClaimPatient,
    ClaimProvider,
    ClaimQueue,
    InstallmentPlan,
)
from .claim_diagnosis_code import ClaimDiagnosisCode
from .claim_line_item import ClaimLineItem
from .command import Command
from .compound_medication import CompoundMedication
from .condition import Condition, ConditionCoding
from .coverage import Coverage, EligibilitySummary, Transactor, TransactorAddress, TransactorPhone
from .detected_issue import DetectedIssue, DetectedIssueEvidence
from .device import Device
from .discount import Discount
from .encounter import Encounter
from .facility import Facility
from .goal import Goal
from .imaging import ImagingOrder, ImagingReport, ImagingReview
from .immunization import (
    Immunization,
    ImmunizationCoding,
    ImmunizationStatement,
    ImmunizationStatementCoding,
)
from .invoice import Invoice
from .lab import (
    LabOrder,
    LabOrderReason,
    LabOrderReasonCondition,
    LabPartner,
    LabPartnerTest,
    LabReport,
    LabReview,
    LabTest,
    LabValue,
    LabValueCoding,
)
from .line_item_transaction import (
    LineItemTransfer,
    NewLineItemAdjustment,
    NewLineItemPayment,
)
from .medication import Medication, MedicationCoding
from .medication_history import (
    MedicationHistoryMedication,
    MedicationHistoryMedicationCoding,
    MedicationHistoryResponse,
    MedicationHistoryResponseStatus,
)
from .medication_statement import MedicationStatement
from .message import Message, MessageAttachment, MessageTransmission
from .note import CurrentNoteStateEvent, Note, NoteStateChangeEvent, NoteType
from .observation import (
    Observation,
    ObservationCoding,
    ObservationComponent,
    ObservationComponentCoding,
    ObservationValueCoding,
)
from .organization import Organization
from .patient import (
    Patient,
    PatientAddress,
    PatientContactPoint,
    PatientExternalIdentifier,
    PatientFacilityAddress,
    PatientMetadata,
    PatientSetting,
)
from .patient_consent import (
    PatientConsent,
    PatientConsentCoding,
    PatientConsentRejectionCoding,
)
from .payment_collection import PaymentCollection
from .payor_specific_charge import PayorSpecificCharge
from .posting import (
    BasePosting,
    BaseRemittanceAdvice,
    BulkPatientPosting,
    CoveragePosting,
    PatientPosting,
)
from .practicelocation import PracticeLocation, PracticeLocationAddress, PracticeLocationSetting
from .protocol_current import ProtocolCurrent
from .protocol_override import ProtocolOverride
from .questionnaire import (
    Interview,
    InterviewQuestionnaireMap,
    InterviewQuestionResponse,
    Question,
    Questionnaire,
    QuestionnaireQuestionMap,
    ResponseOption,
    ResponseOptionSet,
)
from .reason_for_visit import ReasonForVisitSettingCoding
from .referral import Referral, ReferralReport, ReferralReview
from .service_provider import ServiceProvider
from .staff import Staff, StaffAddress, StaffContactPoint, StaffLicense, StaffPhoto, StaffRole
from .stop_medication_event import StopMedicationEvent
from .task import Task, TaskComment, TaskLabel, TaskMetadata, TaskTaskLabel
from .team import Team, TeamContactPoint
from .user import CanvasUser

__all__ = __exports__ = (
    "Appointment",
    "AppointmentMetadata",
    "AppointmentExternalIdentifier",
    "AppointmentLabel",
    "AllergyIntolerance",
    "AllergyIntoleranceCoding",
    "Assessment",
    "BannerAlert",
    "BasePosting",
    "BaseRemittanceAdvice",
    "BillingLineItem",
    "BillingLineItemModifier",
    "BusinessLine",
    "BulkPatientPosting",
    "CanvasUser",
    "CareTeamMembership",
    "CareTeamRole",
    "ChargeDescriptionMaster",
    "Claim",
    "ClaimComment",
    "ClaimCoverage",
    "ClaimDiagnosisCode",
    "ClaimLabel",
    "ClaimLineItem",
    "ClaimPatient",
    "ClaimProvider",
    "ClaimQueue",
    "Command",
    "CompoundMedication",
    "Condition",
    "ConditionCoding",
    "Coverage",
    "CoveragePosting",
    "CurrentNoteStateEvent",
    "DetectedIssue",
    "DetectedIssueEvidence",
    "Device",
    "Discount",
    "EligibilitySummary",
    "Encounter",
    "Facility",
    "Goal",
    "ImagingOrder",
    "ImagingReport",
    "ImagingReview",
    "Immunization",
    "ImmunizationCoding",
    "ImmunizationStatement",
    "ImmunizationStatementCoding",
    "InstallmentPlan",
    "Interview",
    "InterviewQuestionnaireMap",
    "InterviewQuestionResponse",
    "Invoice",
    "LabOrder",
    "LabOrderReason",
    "LabOrderReasonCondition",
    "LabPartner",
    "LabPartnerTest",
    "LabReport",
    "LabReview",
    "LabTest",
    "LabValue",
    "LabValueCoding",
    "LineItemTransfer",
    "Medication",
    "MedicationCoding",
    "MedicationHistoryMedication",
    "MedicationHistoryMedicationCoding",
    "MedicationHistoryResponseStatus",
    "MedicationHistoryResponse",
    "MedicationStatement",
    "Message",
    "MessageAttachment",
    "MessageTransmission",
    "NewLineItemAdjustment",
    "NewLineItemPayment",
    "Note",
    "NoteStateChangeEvent",
    "NoteType",
    "Observation",
    "ObservationCoding",
    "ObservationComponent",
    "ObservationComponentCoding",
    "ObservationValueCoding",
    "Organization",
    "Patient",
    "PatientAddress",
    "PatientContactPoint",
    "PatientExternalIdentifier",
    "PatientFacilityAddress",
    "PatientPosting",
    "PatientSetting",
    "PatientMetadata",
    "PatientConsent",
    "PatientConsentCoding",
    "PatientConsentRejectionCoding",
    "PayorSpecificCharge",
    "PaymentCollection",
    "PracticeLocation",
    "PracticeLocationAddress",
    "PracticeLocationSetting",
    "ProtocolCurrent",
    "ProtocolOverride",
    "Question",
    "Questionnaire",
    "QuestionnaireQuestionMap",
    "ReasonForVisitSettingCoding",
    "Referral",
    "ReferralReport",
    "ReferralReview",
    "ResponseOption",
    "ResponseOptionSet",
    "ServiceProvider",
    "Staff",
    "StaffAddress",
    "StaffLicense",
    "StaffPhoto",
    "StaffRole",
    "StaffContactPoint",
    "StopMedicationEvent",
    "Task",
    "TaskComment",
    "TaskLabel",
    "TaskTaskLabel",
    "TaskMetadata",
    "Team",
    "TeamContactPoint",
    "Transactor",
    "TransactorAddress",
    "TransactorPhone",
)



from typing import cast

from django.db import models

from canvas_sdk.v1.data.base import (
    AuditedModel,
    BaseModelManager,
    BaseQuerySet,
    CommittableQuerySetMixin,
    ForPatientQuerySetMixin,
    IdentifiableModel,
    TimeframeLookupQuerySetMixin,
    TimestampedModel,
    ValueSetLookupQuerySet,
)
from canvas_sdk.v1.data.staff import Staff


class LabReportQuerySet(BaseQuerySet, CommittableQuerySetMixin, ForPatientQuerySetMixin):
    """A queryset for lab reports."""

    pass


LabReportManager = BaseModelManager.from_queryset(LabReportQuerySet)


class TransmissionType(models.TextChoices):
    """Choices for transmission types."""

    FAX = "F", "fax"
    HL7 = "H", "hl7"
    MANUAL = "M", "manual"


class LabReport(AuditedModel, IdentifiableModel):
    """A class representing a lab report."""

    class Meta:
        db_table = "canvas_sdk_data_api_labreport_001"

    objects = cast(LabReportQuerySet, LabReportManager())

    review_mode = models.CharField(max_length=2)
    junked = models.BooleanField()
    requires_signature = models.BooleanField()
    assigned_date = models.DateTimeField()
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="lab_reports", null=True
    )
    transmission_type = models.CharField(choices=TransmissionType.choices, max_length=10)
    for_test_only = models.BooleanField()
    external_id = models.CharField(max_length=40)
    version = models.IntegerField()
    requisition_number = models.CharField(max_length=40)
    review = models.ForeignKey(
        "LabReview", on_delete=models.DO_NOTHING, related_name="reports", null=True
    )
    original_date = models.DateTimeField()
    date_performed = models.DateTimeField()
    custom_document_name = models.CharField(max_length=500)


class LabReviewQuerySet(BaseQuerySet, CommittableQuerySetMixin, ForPatientQuerySetMixin):
    """A queryset for lab reviews."""

    pass


LabReviewManager = BaseModelManager.from_queryset(LabReviewQuerySet)


class LabReview(AuditedModel, IdentifiableModel):
    """A class representing a lab review."""

    class Meta:
        db_table = "canvas_sdk_data_api_labreview_001"

    objects = cast(LabReviewQuerySet, LabReviewManager())

    internal_comment = models.TextField()
    message_to_patient = models.CharField(max_length=2048)
    status = models.CharField(max_length=50)
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="lab_reviews", null=True
    )
    patient_communication_method = models.CharField(max_length=30)


class LabValueTimeframeLookupQuerySetMixin(TimeframeLookupQuerySetMixin):
    """A class that adds queryset functionality to filter using timeframes."""

    @property
    def timeframe_filter_field(self) -> str:
        """Returns the field that should be filtered on. Can be overridden for different models."""
        return "report__original_date"


class LabValueQuerySet(ValueSetLookupQuerySet, LabValueTimeframeLookupQuerySetMixin):
    """LabValueQuerySet."""

    pass


class LabValue(TimestampedModel, IdentifiableModel):
    """A class representing a lab value."""

    class Meta:
        db_table = "canvas_sdk_data_api_labvalue_001"

    objects = LabValueQuerySet.as_manager()

    report = models.ForeignKey(
        "LabReport", on_delete=models.DO_NOTHING, related_name="values", null=True
    )
    value = models.TextField()
    units = models.CharField(max_length=30)
    abnormal_flag = models.CharField(max_length=128)
    reference_range = models.CharField(max_length=128)
    low_threshold = models.CharField(max_length=30)
    high_threshold = models.CharField(max_length=30)
    comment = models.TextField()
    observation_status = models.CharField(max_length=24)


class LabValueCoding(TimestampedModel):
    """A class representing a lab value coding."""

    class Meta:
        db_table = "canvas_sdk_data_api_labvaluecoding_001"

    value = models.ForeignKey(
        LabValue, on_delete=models.DO_NOTHING, related_name="codings", null=True
    )
    code = models.CharField(max_length=128)
    name = models.CharField(max_length=256)
    system = models.CharField(max_length=128)


class LabOrder(AuditedModel, IdentifiableModel):
    """A class representing a lab order."""

    class SpecimenCollectionType(models.TextChoices):
        """Choices for specimen collection types."""

        ON_LOCATION = "L", "on location"
        PATIENT_SERVICE_CENTER = "P", "patient service center"
        OTHER = "O", "other"

    class CourtesyCopyType(models.TextChoices):
        """Choices for courtesy copy types."""

        ACCOUNT = "A", "account"
        FAX = "F", "fax"
        PATIENT = "P", "patient"

    class ManualProcessingStatus(models.TextChoices):
        """Choices for manual processing statuses."""

        MANUAL_PROCESSING_STATUS_NEEDS_REVIEW = "NEEDS_REVIEW", "Needs Review"
        MANUAL_PROCESSING_STATUS_IN_PROGRESS = "IN_PROGRESS", "In Progress"
        MANUAL_PROCESSING_STATUS_PROCESSED = "PROCESSED", "Processed"
        MANUAL_PROCESSING_STATUS_FLAGGED = "FLAGGED", "Flagged"

    class Meta:
        db_table = "canvas_sdk_data_api_laborder_001"

    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="lab_orders", null=True
    )
    ontology_lab_partner = models.CharField(max_length=128)

    note = models.ForeignKey("v1.Note", on_delete=models.DO_NOTHING, null=True)
    comment = models.CharField(max_length=128)
    requisition_number = models.CharField(max_length=32)
    is_patient_bill = models.BooleanField(null=True)
    date_ordered = models.DateTimeField()
    fasting_status = models.BooleanField(null=True)
    specimen_collection_type = models.CharField(
        choices=SpecimenCollectionType.choices, null=True, max_length=10
    )
    transmission_type = models.CharField(choices=TransmissionType.choices, null=True, max_length=10)
    courtesy_copy_type = models.CharField(
        choices=CourtesyCopyType.choices, null=True, max_length=10
    )
    courtesy_copy_number = models.CharField(max_length=32)
    courtesy_copy_text = models.CharField(max_length=64)
    ordering_provider = models.ForeignKey(
        Staff, on_delete=models.DO_NOTHING, related_name="lab_orders", null=True
    )
    parent_order = models.ForeignKey("v1.LabOrder", on_delete=models.DO_NOTHING, null=True)
    healthgorilla_id = models.CharField(max_length=40)
    manual_processing_status = models.CharField(
        choices=ManualProcessingStatus.choices, null=True, max_length=16
    )
    manual_processing_comment = models.TextField(null=True)
    labcorp_abn_url = models.URLField()

    reports = models.ManyToManyField("v1.LabReport", through="v1.LabTest")


class LabOrderReason(AuditedModel):
    """A class representing a lab order reason."""

    class LabReasonMode(models.TextChoices):
        """Choices for lab order reasons."""

        MONITOR = "MO", "monitor"
        INVESTIGATE = "IN", "investigate"
        SCREEN_FOR = "SF", "screen for"
        UNKNOWN = "UNK", "unknown"

    class Meta:
        db_table = "canvas_sdk_data_api_laborderreason_001"

    order = models.ForeignKey(
        LabOrder, on_delete=models.DO_NOTHING, related_name="reasons", null=True
    )
    mode = models.CharField(max_length=30, choices=LabReasonMode)


class LabOrderReasonCondition(TimestampedModel):
    """A class representing a lab order reason's condition."""

    class Meta:
        db_table = "canvas_sdk_data_api_laborderreasoncondition_001"

    reason = models.ForeignKey(
        LabOrderReason, on_delete=models.DO_NOTHING, related_name="reason_conditions", null=True
    )
    condition = models.ForeignKey(
        "v1.Condition",
        on_delete=models.DO_NOTHING,
        related_name="lab_order_reason_conditions",
        null=True,
    )


class LabTest(IdentifiableModel):
    """A class representing a lab test."""

    class LabTestOrderStatus(models.TextChoices):
        """Choicees for lab test order statuses."""

        NEW = "NE", "new"
        STAGED_FOR_REQUISITION = "SR", "staged for requisition"
        SENDING = "SE", "sending"
        SENDING_FAILED = "SF", "sending failed"
        PROCESSING = "PR", "processing"
        PROCESSING_FAILED = "PF", "processing failed"
        RECEIVED = "RE", "received"
        REVIEWED = "RV", "reviewed"
        INACTIVE = "IN", "inactive"

    class Meta:
        db_table = "canvas_sdk_data_api_labtest_001"

    ontology_test_name = models.CharField(max_length=512, blank=True, default="")
    ontology_test_code = models.CharField(max_length=512, blank=True, default="")
    status = models.CharField(max_length=30, choices=LabTestOrderStatus.choices)
    report = models.ForeignKey(
        LabReport, on_delete=models.DO_NOTHING, related_name="tests", null=True
    )
    aoe_code = models.CharField(max_length=10, default="")
    procedure_class = models.CharField(max_length=10, default="")
    specimen_type = models.CharField(max_length=26)
    specimen_source_code = models.CharField(max_length=5)
    specimen_source_description = models.CharField(max_length=255)
    specimen_source_coding_system = models.CharField(max_length=5)
    order = models.ForeignKey(
        LabOrder, on_delete=models.DO_NOTHING, related_name="tests", null=True
    )

    def __str__(self) -> str:
        return f"{self.ontology_test_name}: f{self.ontology_test_code}"


class LabPartner(IdentifiableModel):
    """A class representing a lab partner."""

    class Meta:
        db_table = "canvas_sdk_data_lab_partner_001"

    objects: models.Manager["LabPartner"]

    name = models.CharField(max_length=256)
    active = models.BooleanField()
    electronic_ordering_enabled = models.BooleanField()
    keywords = models.TextField()
    default_lab_account_number = models.CharField(max_length=256)


class LabPartnerTest(IdentifiableModel):
    """A class representing a lab partner's test."""

    class Meta:
        db_table = "canvas_sdk_data_lab_partner_test_001"

    objects: models.Manager["LabPartnerTest"]

    lab_partner = models.ForeignKey(
        "LabPartner", on_delete=models.DO_NOTHING, related_name="available_tests"
    )
    order_code = models.CharField(max_length=256, blank=True)
    order_name = models.TextField()
    keywords = models.TextField(blank=True)
    cpt_code = models.CharField(max_length=256, blank=True, null=True)


__exports__ = (
    "TransmissionType",
    "LabReport",
    "LabReviewQuerySet",
    "LabReview",
    "LabValue",
    "LabValueCoding",
    "LabOrder",
    "LabOrderReason",
    "LabOrderReasonCondition",
    "LabTest",
    "LabPartner",
    "LabPartnerTest",
)



from typing import Literal

from pydantic import Field
from pydantic_core import InitErrorDetails

from canvas_sdk.commands.base import _BaseCommand as BaseCommand
from canvas_sdk.v1.data import Medication, Note


class ChangeMedicationCommand(BaseCommand):
    """A class for managing a ChangeMedication command within a specific note."""

    class Meta:
        key = "changeMedication"
        commit_required_fields = ("medication_id",)

    medication_id: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "medication"}
    )
    sig: str | None = None

    def _get_error_details(
        self, method: Literal["originate", "edit", "delete", "commit", "enter_in_error"]
    ) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        note = None

        if self.note_uuid:
            note = Note.objects.filter(id=self.note_uuid).first()

            if not note:
                errors.append(
                    self._create_error_detail(
                        "value",
                        f"note with id {self.note_uuid} not found.",
                        self.note_uuid,
                    )
                )

        if self.medication_id and note:
            medication = Medication.objects.filter(
                id=self.medication_id, patient=note.patient
            ).first()

            if not medication:
                errors.append(
                    self._create_error_detail(
                        "value",
                        f"Medication with Id {self.medication_id} not found or not associated with the patient.",
                        self.medication_id,
                    )
                )

        return errors


__exports__ = ("ChangeMedicationCommand",)



from functools import cache, lru_cache
from pathlib import Path
from typing import Any

from django.template.backends.django import get_installed_libraries
from django.template.engine import Engine

from canvas_sdk.utils.plugins import plugin_context


@cache
def _installed_template_libraries() -> dict[str, str]:
    """Cache Django's template tag libraries lookup."""
    return get_installed_libraries()


@lru_cache(maxsize=5)
def _engine_for_plugin(plugin_dir: str) -> Engine:
    """Create a Django template engine for the given plugin directory."""
    return Engine(dirs=[plugin_dir], libraries=_installed_template_libraries())


@plugin_context
def render_to_string(
    template_name: str,
    context: dict[str, Any] | None = None,
    **kwargs: Any,
) -> str | None:
    """Load a template and render it with the given context.

    Args:
        template_name (str): The path to the template file, relative to the plugin package.
            If the path starts with a forward slash ("/"), it will be stripped during resolution.
        context (dict[str, Any] | None): A dictionary of variables to pass to the template
            for rendering. Defaults to None, which uses an empty context.
        kwargs (Any): Additional keyword arguments.

    Returns:
        str: The rendered template as a string.

    Raises:
        FileNotFoundError: If the template file does not exist within the plugin's directory
            or if the resolved path is invalid.
    """
    plugin_dir = kwargs["plugin_dir"]
    template_path = Path(plugin_dir / template_name.lstrip("/")).resolve()

    if not template_path.is_relative_to(plugin_dir):
        raise PermissionError(f"Invalid template '{template_name}'")
    elif not template_path.exists():
        raise FileNotFoundError(f"Template {template_name} not found.")

    engine = _engine_for_plugin(plugin_dir)
    return engine.render_to_string(str(template_path), context=context)


__exports__ = ("render_to_string",)



from functools import cached_property
from typing import Any

from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand
from canvas_sdk.commands.commands.questionnaire.question import (
    BaseQuestion,
    CheckboxQuestion,
    IntegerQuestion,
    RadioQuestion,
    ResponseOption,
    TextQuestion,
)
from canvas_sdk.v1.data import Command, Questionnaire

QUESTION_CLASSES: dict[str, type[BaseQuestion]] = {
    ResponseOption.TYPE_TEXT: TextQuestion,
    ResponseOption.TYPE_INTEGER: IntegerQuestion,
    ResponseOption.TYPE_RADIO: RadioQuestion,
    ResponseOption.TYPE_CHECKBOX: CheckboxQuestion,
}


class QuestionnaireCommand(_BaseCommand):
    """A class for managing a Questionnaire command within a specific note."""

    class Meta:
        key = "questionnaire"
        commit_required_fields = ("questionnaire_id",)

    questionnaire_id: str | None = Field(
        default=None, json_schema_extra={"commands_api_name": "questionnaire"}
    )
    result: str | None = None

    @cached_property
    def _questionnaire(self) -> Questionnaire | None:
        if not self.questionnaire_id:
            if command_uuid := self.command_uuid:
                # If the questionnaire is not set, try to fetch it from the command
                try:
                    command_data = Command.objects.values_list("data", flat=True).get(
                        id=command_uuid
                    )
                    if questionnaire_dbid := command_data.get("questionnaire", {}).get("value"):
                        questionnaire = Questionnaire.objects.get(dbid=questionnaire_dbid)
                        self.questionnaire_id = str(questionnaire.id)
                        return questionnaire
                except (Command.DoesNotExist, Questionnaire.DoesNotExist):
                    return None
            else:
                return None

        return Questionnaire.objects.get(id=self.questionnaire_id)  # type: ignore[misc]

    @cached_property
    def questions(self) -> list[BaseQuestion]:
        """
        Returns a list of question objects.

        For each question in the questionnaire, creates an instance of the
        appropriate question subclass based on the question.response_option_set.type.
        """
        question_objs: list[BaseQuestion] = []
        if not self._questionnaire:
            return question_objs

        for question in self._questionnaire.questions.all():
            qdata: dict[str, Any] = {
                "id": question.pk,
                "name": f"question-{question.pk}",
                "label": question.name,
                "coding": {
                    "system": question.code_system,
                    "code": question.code,
                },
                "options": [
                    ResponseOption(
                        dbid=option.pk, name=option.name, code=option.code, value=option.value
                    )
                    for option in question.response_option_set.options.all()
                ]
                if question.response_option_set
                else [],
            }
            q_type = question.response_option_set.type if question.response_option_set else None
            if q_type in QUESTION_CLASSES:
                question_objs.append(QUESTION_CLASSES[q_type](**qdata))
            else:
                raise ValueError(f"Unsupported question type: {q_type}")
        return question_objs

    @property
    def values(self) -> dict:
        """Return the values for the command.

        For questionnaire-related commands, this includes the responses to the questions.
        """
        values = super().values

        values["questions"] = {q.name: q.response for q in self.questions if q.response is not None}

        return values


__exports__ = ("QUESTION_CLASSES", "QuestionnaireCommand")



from django.db import models

from canvas_sdk.v1.data.base import IdentifiableModel


class BusinessLineState(models.TextChoices):
    """BusinessLineStatus."""

    STATE_SUCCESS = "success", "Success"
    STATE_PENDING = "pending", "Pending"
    STATE_ERROR = "error", "Deleted"


class BusinessLine(IdentifiableModel):
    """Business Line."""

    class Meta:
        db_table = "canvas_sdk_data_api_businessline_001"

    name = models.CharField(max_length=255)
    description = models.TextField()
    area_code = models.CharField(max_length=3)
    subdomain = models.CharField(max_length=100)
    active = models.BooleanField()
    state = models.CharField(max_length=20, choices=BusinessLineState.choices)
    organization = models.ForeignKey(
        "v1.Organization", on_delete=models.DO_NOTHING, related_name="business_lines"
    )


__exports__ = (
    "BusinessLineState",
    "BusinessLine",
)



import random
import string
import uuid
from collections.abc import Sequence
from decimal import Decimal


def quantize(deci: Decimal | float | str | tuple[int, Sequence[int], int]) -> Decimal:
    """Rounds a Decimal value to two decimal places."""
    return Decimal(deci).quantize(Decimal(".01"))


def create_key() -> str:
    """Generates a unique key using UUID4."""
    return uuid.uuid4().hex


def generate_mrn(length: int = 9, max_attempts: int = 100) -> str:
    """Generates a unique Medical Record Number (MRN) of specified length."""
    from canvas_sdk.v1.data import Patient

    digits = string.digits

    for _ in range(max_attempts):
        mrn = "".join(random.choices(digits, k=length))
        if not Patient.objects.filter(mrn=mrn).exists():
            return mrn

    raise RuntimeError(f"Unable to generate a unique MRN after {max_attempts} attempts")


def empty_note_body() -> list[dict[str, str]]:
    """Generates an empty note body with 15 empty text elements."""
    return [{"type": "text", "value": ""}] * 15


__exports__ = ()



import factory

from canvas_sdk.v1.data import CanvasUser


class CanvasUserFactory(factory.django.DjangoModelFactory[CanvasUser]):
    """Factory for creating a CanvasUser."""

    class Meta:
        model = CanvasUser

    email = factory.Faker("email")
    phone_number = factory.Faker("phone_number")



from ..value_set import ValueSet


class Ethnicity(ValueSet):
    """

    **Clinical Focus:**

    **Data Element Scope:**

    **Inclusion Criteria:**

    **Exclusion Criteria:**

    ** Used in:** CMS131v14
    """

    VALUE_SET_NAME = "Ethnicity"
    OID = "2.16.840.1.114222.4.11.837"
    DEFINITION_VERSION = "20121025"
    EXPANSION_VERSION = "eCQM Update 2025-05-08"

    CDCREC = {
        "2135-2",  # Hispanic or Latino
        "2186-5",  # Not Hispanic or Latino
    }


class FederalAdministrativeSex(ValueSet):
    """

    **Clinical Focus:** SNOMED CT representations of sex values aligned with US federal administrative requirements across all systems, including quality measure use

    **Data Element Scope:** Patient characteristics

    **Inclusion Criteria:** Limited set of SNOMED CT representations of sex values.

    **Exclusion Criteria:** 

    ** Used in:** CMS131v14
    """

    VALUE_SET_NAME = "Federal Administrative Sex"
    OID = "2.16.840.1.113762.1.4.1021.121"
    DEFINITION_VERSION = "20250228"
    EXPANSION_VERSION = "eCQM Update 2025-05-08"

    SNOMEDCT = {
        "248152002",  # Female (finding)
        "248153007",  # Male (finding)
    }


class PayerType(ValueSet):
    """
    **Clinical Focus:** Categories of types of health care payer entities as defined by the US Public Health Data Consortium SOP code system

    **Data Element Scope:** @code in CCDA r2.1 template Planned Coverage [act: identifier urn:oid:2.16.840.1.113883.10.20.22.4.129 (open)] DYNAMIC

    **Inclusion Criteria:** All codes in the code system

    **Exclusion Criteria:** None

    ** Used in:** CMS131v14
    """

    VALUE_SET_NAME = "Payer Type"
    OID = "2.16.840.1.114222.4.11.3591"
    DEFINITION_VERSION = "20221118"
    EXPANSION_VERSION = "eCQM Update 2025-05-08"

    SOP = {
        "1",      # MEDICARE
        "2",      # MEDICAID
        "3",      # OTHER GOVERNMENT (Federal/State/Local) (excluding Department of Corrections)
        "4",      # DEPARTMENTS OF CORRECTIONS
        "5",      # PRIVATE HEALTH INSURANCE
        "6",      # BLUE CROSS/BLUE SHIELD
        "7",      # MANAGED CARE, UNSPECIFIED (cannot distinguish public from private)
        "8",      # NO PAYMENT from an Organization/Agency/Program/Private Payer Listed
        "9",      # MISCELLANEOUS/OTHER
        "11",     # Medicare Managed Care (Includes Medicare Advantage Plans)
        "12",     # Medicare (Non-managed Care)
        "13",     # Medicare Hospice
        "14",     # Dual Eligibility Medicare/Medicaid Organization
        "19",     # Medicare Other
        "21",     # Medicaid (Managed Care)
        "22",     # Medicaid (Non-managed Care Plan)
        "23",     # Medicaid/SCHIP
        "25",     # Medicaid - Out of State
        "26",     # Medicaid - Long Term Care
        "29",     # Medicaid Other
        "31",     # Department of Defense
        "32",     # Department of Veterans Affairs
        "33",     # Indian Health Service or Tribe
        "34",     # HRSA Program
        "35",     # Black Lung
        "36",     # State Government
        "37",     # Local Government
        "38",     # Other Government (Federal, State, Local not specified)
        "39",     # Other Federal
        "41",     # Corrections Federal
        "42",     # Corrections State
        "43",     # Corrections Local
        "44",     # Corrections Unknown Level
        "51",     # Managed Care (Private)
        "52",     # Private Health Insurance - Indemnity
        "53",     # Managed Care (private) or private health insurance (indemnity), not otherwise specified
        "54",     # Organized Delivery System
        "55",     # Small Employer Purchasing Group
        "56",     # Specialized Stand-Alone Plan
        "59",     # Other Private Insurance
        "61",     # BC Managed Care
        "62",     # BC Insurance Indemnity
        "71",     # HMO (Managed Care, Unspecified)
        "72",     # PPO (Managed Care, Unspecified)
        "73",     # POS (Managed Care, Unspecified)
        "79",     # Other Managed Care
        "81",     # Self-pay (includes applicants for insurance and Medicaid applicants)
        "82",     # No Charge
        "83",     # Refusal to Pay/Bad Debt
        "84",     # Hill Burton Free Care
        "85",     # Research/Donor
        "89",     # No Payment, Other
        "91",     # Foreign National
        "92",     # Other (Non-government)
        "93",     # Disability Insurance
        "94",     # Long-term Care Insurance
        "95",     # Worker's Compensation
        "96",     # Auto Insurance (includes no fault)
        "97",     # Legal Liability / Liability Insurance
        "98",     # Other specified but not otherwise classifiable (includes Hospice - Unspecified plan)
        "99",     # No Typology Code available for payment source
        "111",    # Medicare HMO
        "112",    # Medicare PPO
        "113",    # Medicare POS
        "119",    # Medicare Managed Care Other
        "121",    # Medicare FFS
        "122",    # Medicare Drug Benefit
        "123",    # Medicare Medical Savings Account (MSA)
        "129",    # Medicare Non-managed Care Other
        "141",    # Dual Eligible Special Needs Plan (D-SNP)
        "142",    # Fully Integrated Dual Eligible Special Needs Plan (FIDE-SNP)
        "191",    # Medicare Pharmacy Benefit Manager
        "211",    # Medicaid HMO
        "212",    # Medicaid PPO
        "213",    # Medicaid PCCM (Primary Care Case Management)
        "219",    # Medicaid Managed Care Other
        "291",    # Medicaid Pharmacy Benefit Manager
        "299",    # Medicaid - Dental
        "311",    # TRICARE (CHAMPUS)
        "312",    # Military Treatment Facility
        "313",    # Dental -- Stand Alone
        "321",    # Veteran care - Care provided to Veterans
        "322",    # Non-veteran care
        "331",    # Indian Health Service - Regular
        "332",    # Indian Health Service - Contract
        "333",    # Indian Health Service - Managed Care
        "334",    # Indian Tribe - Sponsored Coverage
        "341",    # Title V (MCH Block Grant)
        "342",    # Migrant Health Program
        "343",    # Ryan White Act
        "344",    # Disaster-related (includes Covid-19)
        "349",    # Other (HRSA Program subcategory)
        "361",    # State SCHIP program (codes for individual states)
        "362",    # Specific state programs (list/local code)
        "369",    # State, not otherwise specified (other state)
        "371",    # Local - Managed care
        "372",    # FFS/Indemnity (Local Government)
        "379",    # Local, not otherwise specified (other local, county)
        "381",    # Federal, State, Local not specified - managed care
        "382",    # Federal, State, Local not specified - FFS
        "389",    # Federal, State, Local not specified - Other
        "391",    # Federal Employee Health Plan - use when known
        "511",    # Commercial Managed Care - HMO
        "512",    # Commercial Managed Care - PPO
        "513",    # Commercial Managed Care - POS
        "514",    # Exclusive Provider Organization
        "515",    # Gatekeeper PPO (GPPO)
        "516",    # Commercial Managed Care - Pharmacy Benefit Manager
        "517",    # Commercial Managed Care - Dental
        "519",    # Managed Care, Other (non HMO)
        "521",    # Commercial Indemnity
        "522",    # Self-insured (ERISA) Administrative Services Only (ASO) plan
        "523",    # Medicare supplemental policy (as second payer)
        "524",    # Indemnity Insurance - Dental
        "529",    # Private health insurance -- other commercial indemnity
        "561",    # Dental (specialized stand-alone plan)
        "562",    # Vision (specialized stand-alone plan)
        "611",    # BC Managed Care - HMO
        "612",    # BC Managed Care - PPO
        "613",    # BC Managed Care - POS
        "614",    # BC Managed Care - Dental
        "619",    # BC Managed Care - Other
        "621",    # BC Indemnity
        "622",    # BC Self-insured (ERISA) Administrative Services Only (ASO) plan
        "623",    # BC Medicare Supplemental Plan
        "821",    # Charity
        "822",    # Professional Courtesy
        "823",    # Research/Clinical Trial
        "951",    # Worker's Comp HMO
        "953",    # Worker's Comp Fee-for-Service
        "954",    # Worker's Comp Other Managed Care
        "959",    # Worker's Comp, Other unspecified
        "1111",   # Medicare Chronic Condition Special Needs Plan (C-SNP)
        "1112",   # Medicare Institutional Special Needs Plan (I-SNP)
        "3111",   # TRICARE Prime -- HMO
        "3112",   # TRICARE Extra -- PPO
        "3113",   # TRICARE Standard - Fee For Service
        "3114",   # TRICARE For Life -- Medicare Supplement
        "3115",   # TRICARE Reserve Select
        "3116",   # Uniformed Services Family Health Plan (USFHP) -- HMO
        "3119",   # Department of Defense - (other)
        "3121",   # Enrolled Prime -- HMO
        "3122",   # Non-enrolled Space Available
        "3123",   # TRICARE For Life (TFL)
        "3211",   # Direct Care - Care provided in VA facilities
        "3212",   # Indirect Care - Care provided outside VA facilities
        "3221",   # Civilian Health and Medical Program for the VA (CHAMPVA)
        "3222",   # Spina Bifida Health Care Program (SB)
        "3223",   # Children of Women Vietnam Veterans (CWVV)
        "3229",   # Other non-veteran care
        "3711",   # HMO (Local - Managed care)
        "3712",   # PPO (Local - Managed care)
        "3713",   # POS (Local - Managed care)
        "3811",   # Federal/State/Local not specified - HMO
        "3812",   # Federal/State/Local not specified - PPO
        "3813",   # Federal/State/Local not specified - POS
        "3819",   # Federal/State/Local not specified - not specified managed care
        "9999",   # Unavailable / No Payer Specified / Blank
        "32121",  # Fee Basis (VA indirect care)
        "32122",  # Foreign Fee/Foreign Medical Program (FMP)
        "32123",  # Contract Nursing Home/Community Nursing Home
        "32124",  # State Veterans Home
        "32125",  # Sharing Agreements
        "32126",  # Other Federal Agency
        "32127",  # Dental Care (VA)
        "32128",  # Vision Care (VA)
    }


class Race(ValueSet):
    """
    **Clinical Focus:** 

    **Data Element Scope:** 

    **Inclusion Criteria:** 

    **Exclusion Criteria:** 

    ** Used in:** CMS131v14
    """

    VALUE_SET_NAME = "Race"
    OID = "2.16.840.1.114222.4.11.836"
    DEFINITION_VERSION = "20121025"
    EXPANSION_VERSION = "eCQM Update 2025-05-08"

    CDCREC = {
        "10025",  # American Indian or Alaska Native
        "20289",  # Asian
        "20545",  # Black or African American
        "20768",  # Native Hawaiian or Other Pacific Islander
        "21063",  # White
        "21311",  # Other Race
    }

class MedicareAdvantagePayer(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to identify concepts for Medicare Advantage payer types.

    **Data Element Scope:** This value set may use a model element related to Patient Characteristic.

    **Inclusion Criteria:** Includes concepts that identify a patient with Medicare Advantage as a payer source.

    **Exclusion Criteria:** Excludes concepts that represent Medicare Fee-For-Service as a payer source.
    """

    VALUE_SET_NAME = "Medicare Advantage Payer"
    OID = "2.16.840.1.113762.1.4.1104.12"
    DEFINITION_VERSION = "20230214"
    EXPANSION_VERSION = "eCQM Update 2025-05-08"

    SOP = {
        "11",  # Medicare Managed Care (Includes Medicare Advantage Plans)
        "111",  # Medicare HMO
        "1111",  # Medicare Chronic Condition Special Needs Plan (C-SNP)
        "1112",  # Medicare Institutional Special Needs Plan (I-SNP)
        "112",  # Medicare PPO
        "113",  # Medicare POS
        "119",  # Medicare Managed Care Other
        "123",  # Medicare Medical Savings Account (MSA)
        "13",  # Medicare Hospice
        "14",  # Dual Eligibility Medicare/Medicaid Organization
        "141",  # Dual Eligible Special Needs Plan (D-SNP)
        "142",  # Fully Integrated Dual Eligible Special Needs Plan (FIDE-SNP)
        "19",  # Medicare Other
        "191",  # Medicare Pharmacy Benefit Manager
    }

class MedicareFfsPayer(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to identify concepts for Medicare Fee-For-Service payer types.

    **Data Element Scope:** This value set may use a model element related to Patient Characteristic.

    **Inclusion Criteria:** Includes concepts that identify a patient with Medicare Fee-For-Service as a payer source.

    **Exclusion Criteria:** Excludes concepts that represent non-Medicare Fee-For-Service as a payer source.
    """

    VALUE_SET_NAME = "Medicare FFS Payer"
    OID = "2.16.840.1.113762.1.4.1104.10"
    DEFINITION_VERSION = "20230214"
    EXPANSION_VERSION = "eCQM Update 2025-05-08"

    SOP = {
        "1",  # MEDICARE
        "12",  # Medicare (Non-managed Care)
        "121",  # Medicare FFS
        "122",  # Medicare Drug Benefit
        "129",  # Medicare Non-managed Care Other
        "13",  # Medicare Hospice
        "14",  # Dual Eligibility Medicare/Medicaid Organization
        "141",  # Dual Eligible Special Needs Plan (D-SNP)
        "142",  # Fully Integrated Dual Eligible Special Needs Plan (FIDE-SNP)
        "19",  # Medicare Other
        "191",  # Medicare Pharmacy Benefit Manager
    }

__exports__ = (
    "Ethnicity",
    "FederalAdministrativeSex",
    "PayerType",
    "Race",
    "MedicareAdvantagePayer",
    "MedicareFfsPayer",
)



from typing import Any

from pydantic_core import InitErrorDetails

from canvas_sdk.effects.metadata import BaseMetadata
from canvas_sdk.v1.data import Patient


class PatientMetadata(BaseMetadata):
    """Effect to upsert a Patient Metadata record."""

    class Meta:
        effect_type = "PATIENT_METADATA"

    patient_id: str

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        if not Patient.objects.filter(id=self.patient_id).exists():
            errors.append(
                self._create_error_detail(
                    "patient_id",
                    f"Patient with id: {self.patient_id} does not exist.",
                    self.patient_id,
                )
            )

        return errors


__exports__ = ("PatientMetadata",)



from pydantic import Field

from canvas_sdk.commands.base import _BaseCommand as BaseCommand


class RemoveAllergyCommand(BaseCommand):
    """A class for managing a Remove Allergy command within a specific note."""

    class Meta:
        key = "removeAllergy"

    allergy_id: str | None = Field(
        description="The external ID of the allergy to remove.",
        default=None,
        json_schema_extra={"commands_api_name": "allergy"},
    )
    narrative: str | None = None


__exports__ = ("RemoveAllergyCommand",)



import functools
import json
from collections.abc import Callable
from pathlib import Path
from typing import Any, TypeVar

import typer

from canvas_cli.utils.print import print

F = TypeVar("F", bound=Callable)


class CLIContext:
    """Class that handles configuration across the CLI.

    Includes methods for:
    * Loading a JSON file with configuration keys into memory.
    * Making a property transient (default, value is not persisted) or persistent (via decorators)

    Loading from a file is dynamic. I.e., you only need to create a new property,
    and mark its setter as persistent if needed.
    """

    # Path to the config file
    _config_file_path: Path

    # Dict with the config file values
    _config_file: dict[str, Any]

    # Base dir for the paths inside the project, not configurable
    _base_dir = Path(__file__).parent / ".." / ".."

    # Base dir for every template used, not configurable
    _base_template_dir: Path = _base_dir / "templates"

    # Dir where the plugin templates are located
    _plugin_template_dir: Path = _base_template_dir / "plugins"

    # Default plugin template name
    _default_plugin_template_name: str = "default"

    # The default host to use for requests
    _default_host: str | None = None

    # When the most recently requested api_token will expire
    _token_expiration_date: str | None = None

    @staticmethod
    def persistent(fn: Callable[..., Any] | None = None, **options: Any) -> Callable[..., Any]:
        """A decorator to store a config value in the file everytime it's changed."""

        def _decorator(fn: Callable[..., Any]) -> Callable[..., Any]:
            @functools.wraps(fn)
            def wrapper(self: "CLIContext", *args: Any, **kwargs: Any) -> Any:
                result = fn(self, *args, **kwargs)
                if args:  # Only store if there are arguments (setter case)
                    value = args[0]
                    print(f"Storing {fn.__name__}={value} in the config file")
                    self._config_file[fn.__name__] = str(value)
                    with open(self._config_file_path, "w") as f:
                        json.dump(self._config_file, f)
                return result

            return wrapper

        return _decorator(fn) if fn else _decorator

    @property
    def plugin_template_dir(self) -> Path:
        """Default Path to use for Plugin templates."""
        return self._plugin_template_dir

    @plugin_template_dir.setter
    @persistent
    def plugin_template_dir(self, new_plugin_template_dir: Path) -> None:
        self._plugin_template_dir = new_plugin_template_dir

    @property
    def default_plugin_template_name(self) -> str:
        """Default template to be used when creating a Plugin."""
        return self._default_plugin_template_name

    @default_plugin_template_name.setter
    @persistent
    def default_plugin_template_name(self, new_default_plugin_template_name: str) -> None:
        self._default_plugin_template_name = new_default_plugin_template_name

    @property
    def default_host(self) -> str | None:
        """Default host to be used when connecting to instances."""
        return self._default_host

    @default_host.setter
    @persistent
    def default_host(self, new_default_host: str | None) -> None:
        self._default_host = new_default_host

    @property
    def token_expiration_date(self) -> str | None:
        """When the most recently requested api_token will expire."""
        return self._token_expiration_date

    @token_expiration_date.setter
    @persistent
    def token_expiration_date(self, new_token_expiration_date: str) -> None:
        self._token_expiration_date = new_token_expiration_date

    def load_from_file(self, file: Path) -> None:
        """Load the given config file into a dict. Aborts execution if it can't decode the file.

        Args:
            file: Path to the JSON config file
        """
        try:
            self._config_file = json.load(file.open("rb"))
        except json.JSONDecodeError:
            print.json(
                "There was a problem loading the config file, please ensure it's valid JSON",
                success=False,
                path=str(file),
            )
            raise typer.Abort() from None

        self._config_file_path = file

        # Get all properties from the class that start with a single underscore,
        # and initialize it with the corresponding key from the config dictionary.
        # This generator assumes the config-file key and property will have the same name (minus the `_`)
        properties = [
            property
            for property in dir(self)
            if property not in ("_config_file", "_config_file_path")
            and not callable(getattr(self, property))
            and not property.startswith("__")
            and property.startswith("_")
        ]

        for property in properties:
            json_var = property[1:]
            if config_value := self._config_file.get(json_var):
                setattr(self, property, config_value)

    def print_config(self) -> None:
        """Print the currently loaded configuration."""
        print.json(message=None, config_file=self._config_file, path=str(self._config_file_path))


# CLIContext pseudo-singleton instance
context: CLIContext = CLIContext()



from django.db import models

from canvas_sdk.v1.data.base import TimestampedModel
from canvas_sdk.v1.data.common import TaxIDType


class Organization(TimestampedModel):
    """Organization."""

    class Meta:
        db_table = "canvas_sdk_data_api_organization_001"

    full_name = models.CharField(max_length=255)
    short_name = models.CharField(max_length=255)
    subdomain = models.CharField(max_length=100)
    logo_url = models.CharField(max_length=255)
    background_image_url = models.CharField(max_length=255)
    background_gradient = models.CharField(max_length=255)
    active = models.BooleanField(default=True)
    tax_id = models.CharField(null=True, max_length=25)
    tax_id_type = models.CharField(choices=TaxIDType.choices, max_length=1)
    group_npi_number = models.CharField(max_length=10)
    group_taxonomy_number = models.CharField(max_length=10)
    include_zz_qualifier = models.BooleanField(default=False)
    main_location = models.OneToOneField(
        "PracticeLocation", related_name="+", null=True, on_delete=models.PROTECT
    )


__exports__ = ("Organization",)
